<?php

namespace Drupal\Tests\term_reference_change\Kernel;

use Drupal\field\Tests\EntityReference\EntityReferenceTestTrait;
use Drupal\KernelTests\KernelTestBase;
use Drupal\node\Entity\Node;
use Drupal\node\NodeInterface;
use Drupal\taxonomy\TermInterface;
use Drupal\term_reference_change\ReferenceMigrator;
use Drupal\Tests\node\Traits\ContentTypeCreationTrait;
use Drupal\Tests\node\Traits\NodeCreationTrait;
use Drupal\Tests\taxonomy\Functional\TaxonomyTestTrait;

/**
 * @group term_reference_change
 */
class ReferenceMigratorTest extends KernelTestBase {

  use TaxonomyTestTrait;
  use NodeCreationTrait;
  use ContentTypeCreationTrait;
  use EntityReferenceTestTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'filter',
    'field',
    'node',
    'taxonomy',
    'term_reference_change',
    'text',
    'user',
    'system',
  ];

  /**
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  private $entityTypeManager;

  /**
   * @var \Drupal\taxonomy\Entity\Vocabulary
   */
  private $vocabulary;

  /**
   * @var \Drupal\taxonomy\TermStorage
   */
  private $termStorage;

  /**
   * {@inheritdoc}
   */
  public function setUp() {
    parent::setUp();
    $this->installConfig(['filter']);
    $this->installSchema('system', 'sequences');
    $this->installSchema('node', 'node_access');
    $this->installEntitySchema('taxonomy_term');
    $this->installEntitySchema('taxonomy_vocabulary');
    $this->installEntitySchema('user');
    $this->installEntitySchema('node');
    $this->installConfig('node');
    $this->setUpContentType('page', 'field_terms');

    $this->entityTypeManager = \Drupal::entityTypeManager();
    $this->vocabulary = $this->createVocabulary();
    $this->termStorage = $this->entityTypeManager->getStorage('taxonomy_term');
  }

  /**
   * @test
   **/
  public function migratesNodes() {
    $sourceTerm = $this->createTerm($this->vocabulary);
    $targetTerm = $this->createTerm($this->vocabulary);
    $node = $this->createNode(['field_terms' => ['target_id' => $sourceTerm->id()]]);

    $sut = new ReferenceMigrator($this->entityTypeManager, \Drupal::service('term_reference_change.reference_finder'));
    $sut->migrateReference($sourceTerm, $targetTerm);

    $this->assertNodeReferencesTerm($node, $targetTerm);
  }

  /**
   * @test
   **/
  public function onlyMigratesLimitedNodes() {
    $sourceTerm = $this->createTerm($this->vocabulary);
    $targetTerm = $this->createTerm($this->vocabulary);
    $node1 = $this->createNode(['field_terms' => ['target_id' => $sourceTerm->id()]]);
    $node2 = $this->createNode(['field_terms' => ['target_id' => $sourceTerm->id()]]);

    $sut = new ReferenceMigrator($this->entityTypeManager, \Drupal::service('term_reference_change.reference_finder'));
    $sut->migrateReference($sourceTerm, $targetTerm, ['node' => [$node1->id()]]);

    $this->assertNodeReferencesTerm($node1, $targetTerm);
    $this->assertNodeReferencesTerm($node2, $sourceTerm);
  }

  /**
   * @test
   **/
  public function doesNotFailWhenReferenceFieldIsMissing() {
    $this->setUpContentType('article', 'field_category');
    $sourceTerm = $this->createTerm($this->vocabulary);
    $targetTerm = $this->createTerm($this->vocabulary);
    $node1 = $this->createNode(['field_terms' => ['target_id' => $sourceTerm->id()]]);
    $node2 = $this->createNode(['type' => 'article']);

    $sut = new ReferenceMigrator($this->entityTypeManager, \Drupal::service('term_reference_change.reference_finder'));
    $sut->migrateReference($sourceTerm, $targetTerm);

    $this->assertNodeReferencesTerm($node1, $targetTerm);

  }

  /**
   * @test
   **/
  public function doesNotFailWhenReferenceFieldIsEmpty() {
    $this->setUpContentType('article', 'field_category');
    $sourceTerm = $this->createTerm($this->vocabulary);
    $targetTerm = $this->createTerm($this->vocabulary);
    $node1 = $this->createNode(['field_terms' => ['target_id' => $sourceTerm->id()]]);
    $node2 = $this->createNode();

    $sut = new ReferenceMigrator($this->entityTypeManager, \Drupal::service('term_reference_change.reference_finder'));
    $sut->migrateReference($sourceTerm, $targetTerm);

    $this->assertNodeReferencesTerm($node1, $targetTerm);
  }

  /**
   * @param \Drupal\node\NodeInterface $node
   * @param \Drupal\taxonomy\TermInterface $targetTerm
   */
  private function assertNodeReferencesTerm(NodeInterface $node, TermInterface $targetTerm) {
    /** @var Node $loadedNode */
    $loadedNode = $this->entityTypeManager->getStorage('node')
      ->load($node->id());
    $referencedTerms = $loadedNode->field_terms->getValue();
    self::assertCount(1, $referencedTerms);
    $firstReference = reset($referencedTerms);
    self::assertEquals($targetTerm->id(), $firstReference['target_id']);
  }

  private function setUpContentType($bundle, $fieldName) {
    $this->createContentType([
      'type' => $bundle,
      'name' => ucfirst($bundle),
      'display_submitted' => FALSE,
    ]);

    $entityType = 'node';
    $fieldLabel = 'Terms';
    $targetEntityType = 'taxonomy_term';
    $this->createEntityReferenceField($entityType, $bundle, $fieldName, $fieldLabel, $targetEntityType);
  }

}
